from math import prod
from numbers import Number
from typing import List


def normalize_tuple(x):
    if not isinstance(x, tuple):
        return (x,)
    return x


def transpose_shape(shape):
    return [shape[1], shape[0]] + list(shape[2:])


def flatten_dictionary(d):
    for i in getattr(d, "values", lambda: d)():
        if isinstance(i, int):
            yield i
        elif i is not None:
            yield from flatten_dictionary(i)


def conv_ops_count(
    x_shape: List[int],
    w_shape: List[int],
    out_shape: List[int],
    transposed: bool = False,
) -> Number:
    """
    Count operations for convolution. Note only multiplication is
    counted. Computation for addition and bias is ignored.
    Operations for a transposed convolution are calculated as
    operations = (x_shape[2:] * prod(w_shape) * batch_size).
    Args:
        x_shape (list(int)): The input shape before convolution.
        w_shape (list(int)): The filter shape.
        out_shape (list(int)): The output shape after convolution.
        transposed (bool): is the convolution transposed
    Returns:
        int: the number of operations
    """
    batch_size = x_shape[0]
    conv_shape = (x_shape if transposed else out_shape)[2:]
    ops = batch_size * prod(w_shape) * prod(conv_shape)
    return ops



from setuptools import find_packages, setup

__version__ = "0.3.4"

install_requires = [
    "torch>=1.13.1",
]

pkg_name = "torch_operation_counter"
setup(
    name=pkg_name,
    version=__version__,
    install_requires=install_requires,
    packages=find_packages(),
    author="Samir Moustafa",
    author_email="samir.moustafa.97@gmail.com",
    url="https://github.com/SamirMoustafa/torch-operation-counter/",
)



import unittest

from torch.nn import Module
from torch.testing import assert_close
from torch import cuda, device, ones, stack, randint, rand, sparse_coo_tensor, zeros

from torch_operation_counter import OperationsCounterMode


class SparseMatrixModule(Module):
    def __init__(self, row, col, value, size):
        super().__init__()
        self.index = stack([row, col], dim=0)
        self.value = value
        self.size = size

    def forward(self, x):
        y = zeros(self.size, x.size(1), device=x.device)
        for i in range(self.index.size(1)):
            y[self.index[0, i]] += self.value[i] * x[self.index[1, i]]
        return y


class TestSparseMatrixMult(unittest.TestCase):
    def setUp(self):
        self.device = device("cuda" if cuda.is_available() else "cpu")

        size = 1000
        row = randint(3, size=(size // 10,), device=self.device)
        col = randint(3, size=(size // 10,), device=self.device)
        value = rand(size // 10, device=self.device)

        self.size = size
        self.sparse_matrix = sparse_coo_tensor(indices=stack([row, col], dim=0),
                                               values=value,
                                               size=(size, size),
                                               device=self.device,
                                               )
        self.sparse_matrix_module = SparseMatrixModule(row, col, value, size).to(self.device)

    def test_correctness(self):
        x = ones(self.size, 10, device=self.device)
        expected_output = self.sparse_matrix.matmul(x)
        output = self.sparse_matrix_module(x)
        assert_close(output, expected_output)

    def test_num_ops(self):
        x = ones(self.size, 1, device=self.device)
        with OperationsCounterMode() as ops_counter:
            self.sparse_matrix_module(x)
        # # Logging the operations count
        # counter_dict = ops_counter.operations_count
        # for module, operation in counter_dict.items():
        #     for op_name, op_count in operation.items():
        #         print(f"{module}::{op_name}: {op_count} OPs")
        # The number of operations should be 2 * nnz
        expected_ops = 2 * self.sparse_matrix._nnz()
        self.assertEqual(ops_counter.total_operations, expected_ops)


if __name__ == "__main__":
    unittest.main()



__all__ = ["aten_ignore"]

ignore = [
    "_local_scalar_dense",
    "_reshape_alias",
    "argmax",
    "bernoulli_",
    "cat",
    "clone",
    "copy_",
    "detach",
    "dropout",
    "empty",
    "empty_like",
    "expand",
    "expand_as",
    "fill_",
    "flatten",
    "full",
    "full_like",
    "getattr",
    "getitem",
    "index",
    "index_put_",
    "index_select",
    "masked_fill_",
    "new_empty",
    "new_empty",
    "new_full",
    "new_zeros",
    "ones",
    "ones_like",
    "permute",
    "repeat",
    "select",
    "sign",
    "size",
    "slice",
    "squeeze",
    "squeeze",
    "t",
    "type_as",
    "unsqueeze",
    "view",
    "view_as",
    "zeros",
    "zeros_like",
    "eye",
]


aten_ignore = [f"aten.{ignore_i}" for ignore_i in ignore]
